import abc
import asyncio
import datetime
import re
import time
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, AsyncIterator, Self, TypeAlias


class SparkplugBError(Exception):
    """A Sparkplug B norm is broken"""


class Clock(abc.ABC):
    # @abc.abstractmethod
    # def now(self) -> int:
    #     """Return current timestamp in ms"""

    # @abc.abstractmethod
    # def sleep(self, seconds: float):
    #     """Syncronous sleep for several seconds"""

    # @abc.abstractmethod
    # async def asleep(self, milliseconds: int):
    #     """Asyncronous sleep for several seconds"""

    @abc.abstractmethod
    def timestamp(self) -> int:
        """Return current timestamp in ms"""

    @abc.abstractmethod
    def get_ticker(self, interval_ms: int) -> AsyncIterator[int]:
        """Raise every interval"""

    @abc.abstractmethod
    async def wait_for(self, task: asyncio.Task, timeout_ms: int):
        """Wait for finish a task for timeout seconds, otherwise raises TimeoutError"""


class UtcClock(Clock):
    class Ticker:
        def __init__(self, interval_ms: int, clock: Clock):
            self._next_ts = ((clock.timestamp() // interval_ms) + 1) * interval_ms
            self._clock = clock
            self._interval = interval_ms

        def __aiter__(self):
            return self

        async def __anext__(self):
            wait = (self._next_ts - self._clock.timestamp()) / 1000
            if wait <= 0:
                return self._clock.timestamp()
            await asyncio.sleep(wait)
            self._next_ts += self._interval
            return self._clock.timestamp()

    def __str__(self) -> str:
        return datetime.datetime.now(datetime.timezone.utc).isoformat()

    def sleep(self, milli_seconds: int):
        time.sleep(milli_seconds / 1000)

    async def asleep(self, mili_seconds: int):
        await asyncio.sleep(mili_seconds / 1000)

    def timestamp(self) -> int:
        return int(datetime.datetime.now(datetime.timezone.utc).timestamp() * 1000)

    async def wait(
        self, tasks: list[asyncio.Task], timeout: float | None = None
    ) -> tuple[set[asyncio.Task], set[asyncio.Task]]:
        return await asyncio.wait(tasks, timeout=timeout)

    async def wait_for(self, task: asyncio.Task, timeout_ms: int):
        return await asyncio.wait_for(task, timeout=timeout_ms / 1000)

    def get_ticker(self, interval_ms: int) -> AsyncIterator[int]:
        return self.Ticker(interval_ms, self)


@dataclass
class DeviceKey:
    path: str

    def __post_init__(self):
        if self.path.count("/") > 2:
            raise ValueError(
                f"Path {self.path} has too much levels to be a device/host"
            )

    @property
    def node(self) -> Self | None:
        match self.path.count("/"):
            case 1:
                return self
            case 2:
                tokens = self.path.split("/")
                return self.__class__("/".join(tokens[:2]))

    @property
    def hostname(self) -> str:
        if self.path.count("/") == 0:
            return self.path
        return ""

    @property
    def group(self) -> str:
        if self.path.count("/"):
            return self.path[: self.path.find("/")]
        return ""

    @property
    def node_name(self) -> str:
        tokens = self.path.split("/")
        if len(tokens) > 1:
            return tokens[1]
        return ""

    @property
    def device_name(self) -> str:
        tokens = self.path.split("/")
        if len(tokens) == 3:
            return tokens[2]
        return ""

    def __str___(self):
        return self.path

    def __hash__(self):
        return hash(self.path)


class MessageType(Enum):
    NBIRTH = "NBIRTH"
    NDATA = "NDATA"
    NDEATH = "NDEATH"
    NCMD = "NCMD"
    DBIRTH = "DBIRTH"
    DDATA = "DDATA"
    DDEATH = "DDEATH"
    DCMD = "DCMD"
    STATE = "STATE"

    def is_node_message(self) -> bool:
        return self in (
            MessageType.NBIRTH,
            MessageType.NDATA,
            MessageType.NDEATH,
            MessageType.NCMD,
        )

    def is_device_message(self) -> bool:
        return self in (
            MessageType.DBIRTH,
            MessageType.DDATA,
            MessageType.DDEATH,
            MessageType.DCMD,
        )

    def is_host_message(self) -> bool:
        return self == MessageType.STATE


class Quality(Enum):
    BAD = 0
    GOOD = 192
    STALE = 500


ValueType: TypeAlias = bool | int | float | str | bytes | None


class DataType(Enum):
    Unknown = 0
    Int8 = 1
    Int16 = 2
    Int32 = 3
    Int64 = 4
    UInt8 = 5
    UInt16 = 6
    UInt32 = 7
    UInt64 = 8
    Float = 9
    Double = 10
    Boolean = 11
    String = 12
    DateTime = 13
    Text = 14
    UUID = 15
    DataSet = 16
    Bytes = 17
    File = 18
    Template = 19
    PropertySet = 20
    PropertySetList = 21
    Int8Array = 22
    Int16Array = 23
    Int32Array = 24
    Int64Array = 2
    UInt8Array = 26
    UInt16Array = 27
    UInt32Array = 28
    UInt64Array = 29
    FloatArray = 30
    DoubleArray = 31
    BooleanArray = 32
    StringArray = 33
    DateTimeArray = 34

    @classmethod
    def for_(cls, value: ValueType) -> Self:
        v_type = type(value)
        if v_type not in _DEFAULT_DTs:
            raise ValueError(f"There is no default for type {v_type}")
        return _DEFAULT_DTs[v_type]

    def convert_value(self, str_value: str) -> ValueType:
        if str_value == "None":
            return

        if "Int" in self.name:
            return int(str_value)

        if "Float" in self.name or "Double" in self.name:
            return float(str_value)

        if "Boolean" == self.name:
            return str_value.lower() == "true"

        if "Bytes" == self.name:
            return str_value.encode()

        return str_value


_DEFAULT_DTs = {
    int: DataType.Int64,
    float: DataType.Float,
    bool: DataType.Boolean,
    str: DataType.String,
    bytes: DataType.Bytes,
}


@dataclass(frozen=True)
class PropertyValue:
    value: ValueType
    data_type: DataType

    def to_dict(self) -> dict[str, ValueType]:
        return {"value": self.value, "dataType": self.data_type.name}

    @classmethod
    def from_dict(cls, dump: dict[str, Any]) -> Self:
        datatype = (
            DataType[dump["dataType"]]
            if type(dump["dataType"]) is str
            else DataType(dump["dataType"])
        )
        return cls(dump["value"], datatype)

    @classmethod
    def from_value(cls, value: ValueType) -> Self:
        value_type = type(value)
        if value_type not in _DEFAULT_DTs:
            raise ValueError(f"Not default for type {type(value)}")

        datatype = _DEFAULT_DTs[value_type]

        return cls(value, datatype)


# @dataclass(frozen=True)
# class PropertySet:
#     """Map or properties"""

#     keys: tuple[str, ...] = field(default_factory=tuple)
#     values: tuple[PropertyValue, ...] = field(default_factory=tuple)

#     def __contains__(self, key: str) -> bool:
#         return key in self.keys

#     def __getitem__(self, key: str) -> PropertyValue:
#         try:
#             index = self.keys.index(key)
#         except ValueError:
#             raise KeyError(f"Key {key} not found in properties")

#         return self.values[index]

#     def __bool__(self) -> bool:
#         return bool(self.keys)

#     def get(self, key: str, default: ValueType) -> ValueType:
#         try:
#             keys = [key.lower() for key in self.keys]
#             index = keys.index(key.lower())
#         except ValueError:
#             return default
#         return self.values[index].value

#     @classmethod
#     def from_dict(cls, dump: dict[str, Any]) -> Self:
#         """Construct a property set from a dict"""
#         keys = tuple(sorted(dump.keys()))
#         return cls(keys, tuple([PropertyValue.from_dict(dump[key]) for key in keys]))

#     @classmethod
#     def from_kwargs(cls, **kwargs) -> Self:
#         """Construct property set from keywords arguments"""
#         keys = tuple(sorted(kwargs.keys()))

#         return cls(keys, tuple([kwargs[key] for key in keys]))

#     def as_dict(self) -> dict[str, dict[str, Any]]:
#         """Convert object to a dict"""
#         return {key: value.to_dict() for key, value in zip(self.keys, self.values)}


def _to_snake_case(input_string):
    """
    Convert any naming convention to snake_case.
    Handles camelCase, PascalCase, kebab-case, space-separated, and mixed formats.
    """
    if not input_string:
        return input_string

    # Replace special characters (except alphanumeric and spaces) with spaces
    s = re.sub(r"[^a-zA-Z0-9\s]", " ", input_string)

    # Convert camelCase/PascalCase by adding spaces before uppercase letters
    s = re.sub(r"([a-z0-9])([A-Z])", r"\1 \2", s)

    # Convert to lowercase and replace multiple spaces with a single space
    s = s.lower().strip()
    s = re.sub(r"\s+", " ", s)

    # Replace spaces with underscores
    s = s.replace(" ", "_")

    # Handle multiple underscores
    s = re.sub(r"_+", "_", s)

    # Remove leading/trailing underscores
    s = s.strip("_")

    return s


class PropertySet(dict):
    """
    A dictionary subclass that converts keys to snake_case before storing
    and allows searching with any key notation.
    """

    def __setitem__(self, key, value):
        """Store the item with the key converted to snake_case."""
        if type(value) is not PropertyValue:
            raise ValueError(f"Value shall be PropertyValue, not {type(value)}")

        snake_key = _to_snake_case(key)
        super().__setitem__(snake_key, value)

    def __getitem__(self, key):
        """Retrieve an item using any key notation."""
        snake_key = _to_snake_case(key)
        return super().__getitem__(snake_key)

    def __delitem__(self, key):
        """Delete an item using any key notation."""
        snake_key = _to_snake_case(key)
        super().__delitem__(snake_key)

    def __contains__(self, key):
        """Check if a key (in any notation) exists."""
        snake_key = _to_snake_case(key)
        return super().__contains__(snake_key)

    def get(self, key, default=None):
        """Get an item with any key notation, with optional default."""
        snake_key = _to_snake_case(key)
        return super().get(snake_key, default)

    def setdefault(self, key, default=None):
        """Set default value for a key in any notation."""
        snake_key = _to_snake_case(key)
        return super().setdefault(snake_key, default)

    def pop(self, key, *args):
        """Pop an item using any key notation."""
        snake_key = _to_snake_case(key)
        return super().pop(snake_key, *args)

    def update(self, *args, **kwargs):
        """Update the dictionary with key-value pairs, converting keys to snake_case."""
        if args:
            other = args[0]
            if isinstance(other, dict):
                for key, value in other.items():
                    self[key] = value
            else:
                for key, value in other:
                    self[key] = value
        for key, value in kwargs.items():
            self[key] = value


@dataclass
class Metric:
    """Data Transfer Object (dto) to be included in Sparkplug Messages"""

    timestamp: int
    value: ValueType
    data_type: DataType
    alias: int = 0
    name: str = ""
    properties: PropertySet = field(default_factory=PropertySet)
    is_transient: bool = False
    is_historical: bool = False

    @classmethod
    def from_dict(cls, dump: dict[str, Any]) -> Self:
        """Construct dto from a dict"""
        return cls(
            dump["timestamp"],
            dump["value"],
            DataType[dump["dataType"]],
            dump.get("alias", 0),
            dump.get("name", ""),
            PropertySet(
                {
                    key: PropertyValue.from_dict(value)
                    for key, value in dump.get("properties", {}).items()
                }
            ),
            dump.get("is_transient", False),
            dump.get("is_historical", False),
        )

    def as_dict(self) -> dict[str, Any]:
        """Convert dto to dict for parsing"""
        dump = {
            "timestamp": self.timestamp,
            "value": self.value,
            "dataType": self.data_type.name,
        }
        if self.alias:
            dump["alias"] = self.alias

        if self.name:
            dump["name"] = self.name

        if self.is_transient:
            dump["is_transient"] = True

        if self.properties:
            dump["properties"] = {
                key: value.as_dict() for key, value in self.properties
            }

        if self.is_historical:
            dump["is_historical"] = True
        return dump

    @property
    def quality(self) -> Quality | None:
        if not self.properties or "quality" not in self.properties:
            return

        return Quality(self.properties["quality"].value)

    @quality.setter
    def quality(self, q: Quality):
        self.properties["quality"] = PropertyValue(q.value, DataType.Int32)


@dataclass
class HostOptions:
    group: str = "+"
    reorder_time: float = 2.0  # in s
    notify_timeout: float = 10.0  # in s

    def apply(self, session):
        session._opts = self


@dataclass
class NodeOptions:
    historian_dir: str = (
        ""  # default is InMemHistorian. Set directory path to be FSHistorian
    )
    primary: str = ""
    scan_rate: int = 60000  # in ms, default is 1 min
    write_timeout: float = 10.0  # units in s
    reorder_time: float = 2.0  # units in s
    max_payload_size: int = 0  # units are bytes, default 0, no split payload

    def apply(self, session):
        session._opts = self



from dataclasses import dataclass, field
from enum import Enum
from typing import Self

from ..shared import DeviceKey, Metric


@dataclass
class MQTTMessage:
    """Message managed by the MQTT server"""

    topic: str
    payload: bytes
    qos: int = 0
    retain: bool = False


@dataclass
class Seq:
    value: int = -1

    def next(self) -> "Seq":
        if self.value == 255:
            return Seq(0)

        return Seq(self.value + 1)


class Event(Enum):
    NODE_IS_BORN = 0
    DEVICE_IS_BORN = 1
    NODE_DATA_HAS_CHANGED = 2
    DEVICE_DATA_HAS_CHANGED = 3
    DEVICE_HAS_DIED = 4
    NODE_HAS_DIED = 5
    NODE_COMMAND_IS_SENT = 6
    DEVICE_COMMAND_IS_SENT = 7
    HOST_STATE_HAS_CHANGED = 10
    OUT_OF_SEQUENCE = 100

    @classmethod
    def from_message_type(cls, value: str) -> "Event":
        events = {
            "NBIRTH": Event.NODE_IS_BORN,
            "NDATA": Event.NODE_DATA_HAS_CHANGED,
            "NDEATH": Event.NODE_HAS_DIED,
            "NCMD": Event.NODE_COMMAND_IS_SENT,
            "DBIRTH": Event.DEVICE_IS_BORN,
            "DDATA": Event.DEVICE_DATA_HAS_CHANGED,
            "DDEATH": Event.DEVICE_HAS_DIED,
            "DCMD": Event.DEVICE_COMMAND_IS_SENT,
            "STATE": Event.HOST_STATE_HAS_CHANGED,
        }
        if value not in events:
            raise ValueError(f'Message type "{value}" is not a registered event')
        return events[value]

    def as_message_type(self) -> str:
        msgs = {
            Event.NODE_IS_BORN: "NBIRTH",
            Event.NODE_DATA_HAS_CHANGED: "NDATA",
            Event.NODE_HAS_DIED: "NDEATH",
            Event.NODE_COMMAND_IS_SENT: "NCMD",
            Event.DEVICE_IS_BORN: "DBIRTH",
            Event.DEVICE_DATA_HAS_CHANGED: "DDATA",
            Event.DEVICE_HAS_DIED: "DDEATH",
            Event.DEVICE_COMMAND_IS_SENT: "DCMD",
            Event.HOST_STATE_HAS_CHANGED: "STATE",
        }
        return msgs[self]


@dataclass
class Payload:
    timestamp: int
    metrics: list[Metric] = field(default_factory=list)
    seq: int | None = None
    online: bool | None = None


@dataclass
class Message:
    event: Event
    key: DeviceKey
    metrics: list[Metric] = field(
        default_factory=list
    )  # No when death certificate or HOST_STATE_HAS_CHANGED
    online: bool | None = None  # No when node/device message
    sending_ts: int = 0  # if it is 0 is because it's not yet sent

    def get_topic(self) -> str:
        if self.event in (
            Event.NODE_IS_BORN,
            Event.NODE_DATA_HAS_CHANGED,
            Event.NODE_HAS_DIED,
            Event.NODE_COMMAND_IS_SENT,
        ):
            return f"spBv1.0/{self.key.group}/{self.event.as_message_type()}/{self.key.node_name}"

        if self.event in (
            Event.DEVICE_IS_BORN,
            Event.DEVICE_DATA_HAS_CHANGED,
            Event.DEVICE_HAS_DIED,
            Event.DEVICE_COMMAND_IS_SENT,
        ):
            return f"spBv1.0/{self.key.group}/{self.event.as_message_type()}/{self.key.node_name}/{self.key.device_name}"

        if self.event == Event.HOST_STATE_HAS_CHANGED:
            return f"spBv1.0/STATE/{self.key.hostname}"

        raise NotImplementedError(f"Event {self.event} has not topic defined")

    def get_payload(self, seq: int | None = None, bd_metric: Metric | None = None):
        # Preconditions
        if bd_metric and self.event not in (Event.NODE_IS_BORN, Event.NODE_HAS_DIED):
            raise ValueError(f"Event {self.event} can not have any bd_metric")

        if seq is not None and self.event in (
            Event.HOST_STATE_HAS_CHANGED,
            Event.DEVICE_COMMAND_IS_SENT,
            Event.NODE_COMMAND_IS_SENT,
            Event.NODE_HAS_DIED,
        ):
            raise ValueError(f"Event {self.event} doesn't allow sequence value {seq}")

        return Payload(
            self.sending_ts,
            self.metrics + [bd_metric] if bd_metric else self.metrics,
            seq,
            self.online,
        )

    @staticmethod
    def _get_event_and_key(topic: str) -> tuple[Event, DeviceKey]:
        tokens = topic.split("/")
        if tokens[0] != "spBv1.0":
            raise ValueError(
                f'First level of topic "{tokens[0]}" is not Sparkplug B compliance'
            )

        if len(tokens) < 3 or len(tokens) > 5:
            raise ValueError(f'Topic "{topic}" has not correct number of levels')

        key = None
        event = None
        match len(tokens):
            case 3:
                key = DeviceKey(tokens[2])
                event = Event.from_message_type(tokens[1])
            case 4:
                key = DeviceKey("/".join([tokens[1], tokens[3]]))
                event = Event.from_message_type(tokens[2])
            case _:
                key = DeviceKey("/".join([tokens[1]] + tokens[3:]))
                event = Event.from_message_type(tokens[2])

        return (event, key)

    @classmethod
    def construct(cls, topic: str, payload: Payload) -> Self:
        event, key = cls._get_event_and_key(topic)

        metrics = payload.metrics
        if event in (Event.NODE_IS_BORN, Event.NODE_HAS_DIED):  # remove bd_metric
            metrics = metrics[:-1]

        return cls(event, key, metrics, payload.online, payload.timestamp)



from copy import deepcopy

from aiospb.hosts import MinMetricInfo, MinMetricInfoCache
from aiospb.nodes import (
    Historian,
    MetricInfo,
    MetricNotFound,
    MetricsNetwork,
    Reading,
    ReadingStore,
)
from aiospb.shared import DataType, DeviceKey, Metric, Quality, ValueType


class InMemHistorian(Historian):
    """Implementation of historian in memory,

    for acceptance tests purpose"""

    def __init__(self):
        self._data = {}

    async def save(self, key: DeviceKey, metrics: list[Metric]):
        if key not in self._data:
            self._data[key] = []

        metrics = deepcopy(metrics)
        for metric in metrics:
            metric.is_historical = True
        self._data[key].extend(metrics)

    async def load(self, key: DeviceKey) -> list[Metric]:
        if key not in self._data:
            return []
        return self._data[key]

    async def clear(self, key: DeviceKey):
        self._data.pop(key)

    def get_device_keys(self, node_key: DeviceKey) -> list[DeviceKey]:
        return [key for key in self._data.keys() if key.node == node_key]


class InMemReadingStore(ReadingStore):
    """Implementation in RAM memory, not valid if too much metrics"""

    def __init__(self):
        self._readings = {}

    async def get(self, aliases: list[int]) -> list[Reading]:
        return [
            self._readings.get(alias, Reading(None, DataType.Unknown, Quality.BAD))
            for alias in aliases
        ]

    async def set(self, readings: dict[int, Reading]):
        self._readings.update(readings)


class InMemMinMetricInfoCache(MinMetricInfoCache):
    """Implementation in RAM memory,

    for low quantity of connected nodes and metrics, otherwise use fs"""

    def __init__(self):
        self._data = {}

    async def get(self, key: DeviceKey, names: list[str]) -> list[MinMetricInfo]:
        if key not in self._data:
            return [MinMetricInfo(DataType.Unknown)] * len(names)

        return [
            self._data[key].get(name, MinMetricInfo(DataType.Unknown)) for name in names
        ]

    async def set(self, key: DeviceKey, info: dict[str, MinMetricInfo]):
        if key not in self._data:
            self._data[key] = {}

        self._data[key].update(info)

    async def remove(self, key: DeviceKey):
        self._data.pop(key, None)



import asyncio
import logging
from dataclasses import dataclass
from typing import AsyncIterator

import aiomqtt
from aiomqtt.exceptions import MqttError as AioMqttError

from aiospb.messages import MQTT5Client, MQTTError, MQTTMessage

logger = logging.getLogger(__name__)


@dataclass
class MQTTConfig:
    hostname: str
    port: int
    username: str = ""
    password: str = ""
    ca_cert: str = ""
    keepalive: int = 30


class PahoMQTTClient(MQTT5Client):
    """Implementation with Paho of MQTT5Publisher and MQTTSubscriver"""

    def __init__(self, client_id: str, config: MQTTConfig):
        self._id = client_id
        self._config = config
        self._client = None

    @property
    def keepalive(self) -> float:
        return self._config.keepalive

    async def connect(self, will: MQTTMessage | None = None):
        will_ = None
        if will:
            will_ = aiomqtt.Will(will.topic, will.payload, will.qos, will.retain)

        tls_pars = None
        if self._config.ca_cert:
            tls_pars = aiomqtt.TLSParameters(ca_certs=self._config.ca_cert)

        self._client = aiomqtt.Client(
            self._config.hostname,
            self._config.port,
            identifier=self._id,
            username=self._config.username,
            password=self._config.password,
            will=will_,
            protocol=aiomqtt.ProtocolVersion.V5,
            tls_params=tls_pars,
            clean_start=True,
            keepalive=self._config.keepalive,
            timeout=30,
        )

        try:
            await asyncio.wait_for(self._client.__aenter__(), timeout=60)
        except AioMqttError as e:
            raise MQTTError(e) from e

    async def publish(self, message: MQTTMessage):
        if self._client is None:
            raise RuntimeError("Client not connected to Mqtt Server")

        attempts = 0
        while True:
            try:
                await self._client.publish(
                    message.topic,
                    message.payload,
                    qos=message.qos,
                    retain=message.retain,
                    timeout=10.0,
                )
                return
            except asyncio.TimeoutError:
                attempts += 1
                logger.warning(
                    f"Failure by timeout publishing to {message.topic}, attempt {attempts}"
                )
                if attempts > 3:
                    raise TimeoutError(
                        f"After 3 attempts not possible to publish to {message.topic}"
                    )
            except AioMqttError as e:
                raise MQTTError(e) from e

    async def disconnect(self):
        if self._client is None:
            return

        try:
            await self._client.__aexit__(None, None, None)
        except AioMqttError as e:
            raise MQTTError(e) from e
        finally:
            self._client = None

    def is_connected(self):
        return self._client is not None and self._client._connected

    async def subscribe(self, wildcard: str, qos: int):
        if self._client is None:
            raise RuntimeError("Mqtt client is not connected")

        try:
            await self._client.subscribe(wildcard, qos=qos)
        except AioMqttError as e:
            raise MQTTError(e) from e

    async def unsubscribe(self, wildcard: str):
        if self._client is None:
            raise RuntimeError("Mqtt client is not connected")

        try:
            await self._client.unsubscribe(wildcard)
        except AioMqttError as e:
            raise MQTTError(e) from e

    @property
    def messages(self) -> AsyncIterator[MQTTMessage]:
        class MQTTMessages:
            def __init__(self, msgs: aiomqtt.MessagesIterator):
                self._msgs = msgs

            def __aiter__(self):
                self._msgs.__aiter__()
                return self

            async def __anext__(self) -> MQTTMessage:
                try:
                    msg = await self._msgs.__anext__()
                    if type(msg.payload) is not bytes:
                        raise TypeError(
                            "Expected payload to be bytes, not {type(msg.payload)}"
                        )
                    return MQTTMessage(
                        msg.topic.value, msg.payload, msg.qos, msg.retain
                    )
                except AioMqttError as e:
                    raise MQTTError(e) from e

        if self._client is None or not self._client._connected:
            raise MQTTError(Exception("Mqtt client is not connected"))

        return MQTTMessages(self._client.messages)



from dataclasses import dataclass, field

from typing import Self
from aiospb.shared import DataType, PropertySet, ValueType, Quality, Metric


@dataclass
class Reading:
    value: ValueType
    data_type: DataType = DataType.Unknown
    quality: Quality = Quality.GOOD

    def update(self, newer: Self) -> "Reading":
        value = newer.value if newer.quality == Quality.GOOD else self.value
        return Reading(value, self.data_type, newer.quality)

    def create_change_metric(
        self, timestamp: int, alias: int, newer: Self
    ) -> Metric | None:
        reading = self.update(newer)
        if reading == self:
            return

        metric = Metric(timestamp, reading.value, reading.data_type, alias, "")
        if reading.quality != self.quality:
            metric.quality = reading.quality

        return metric


@dataclass
class MetricInfo:
    """Stable metric data, stable during a session"""

    name: str
    data_type: DataType
    properties: PropertySet = field(default_factory=PropertySet)
    alias: int = 0
    is_transient: bool = False

    def create_birth_metric(self, timestamp: int, reading: Reading) -> Metric:
        """Create metric from the metric core"""
        metric = Metric(
            timestamp,
            reading.value,
            self.data_type,
            self.alias,
            self.name,
            self.properties,
            is_transient=self.is_transient,
        )

        if reading.quality != Quality.GOOD:
            metric.quality = reading.quality

        return metric



import asyncio
import logging
from collections.abc import Callable
from typing import Coroutine

from aiospb.nodes.ports import DeviceConnectionIsBroken
from aiospb.nodes.values import Reading
from aiospb.shared import Clock, DeviceKey, Metric, Quality

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .devices import MetricsNetwork, ReadingStore


class Scanner:
    def __init__(
        self,
        key: DeviceKey,
        aliases: list[int],
        metrics_net: "MetricsNetwork",
        readings: "ReadingStore",
        callback: Callable[[list[Metric]], Coroutine[None, None, None]],
        clock: Clock,
    ):
        self._key = key
        self._net = metrics_net
        self._readings = readings
        self._callback = callback
        self._clock = clock
        self._aliases = aliases

        self._active_scans = set()
        self._loop_task = None

    async def _scan_loop(
        self,
        scan_rate: int,
    ):
        try:
            async for tick in self._clock.get_ticker(scan_rate):
                if self._active_scans:
                    logger.warning(
                        f"Starting a new scan at {tick}, but not finished {len(self._active_scans)} previous"
                    )

                scan = asyncio.create_task(self._scan(scan_rate), name=f"@{tick}")
                self._active_scans.add(scan)
                scan.add_done_callback(lambda task: self._active_scans.remove(task))

        except asyncio.CancelledError:
            for scan in self._active_scans:
                scan.cancel()
            return

    async def _scan(
        self,
        scan_rate: int,
    ):
        try:  # TODO: Manage self._readings exceptions
            last_values = await self._readings.get(self._aliases)
            tasks = [
                asyncio.create_task(self._read(alias, int(scan_rate * 0.9)))
                for alias in self._aliases
            ]
            await asyncio.wait(tasks)

            metrics = []
            readings = {}
            for last, task, alias in zip(last_values, tasks, self._aliases):
                ts, reading = task.result()
                metric = last.create_change_metric(ts, alias, reading)
                if metric:
                    metrics.append(metric)
                    readings[alias] = last.update(reading)

            if metrics:
                metrics = sorted(metrics, key=lambda m: m.timestamp)
                await self._callback(metrics)
                await self._readings.set(readings)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            logger.error(f"Scanning at {self._clock.timestamp()}")
            logger.exception(e)
            raise e

    async def _read(self, alias: int, timeout_ms: int) -> tuple[int, Reading]:
        try:
            value = await self._clock.wait_for(
                asyncio.create_task(self._net.read(self._key, alias)), timeout_ms
            )
        except TimeoutError:
            return (self._clock.timestamp(), Reading(None, quality=Quality.STALE))
        except DeviceConnectionIsBroken:
            return (self._clock.timestamp(), Reading(None, quality=Quality.STALE))
        except Exception as e:
            logger.error(f"Reading metric with alias {alias}")
            logger.exception(e)
            return (self._clock.timestamp(), Reading(None, quality=Quality.BAD))

        return (self._clock.timestamp(), Reading(value))

    def start(
        self,
        scan_rate: int,
    ):
        if self._loop_task:
            self._loop_task.cancel()

        self._loop_task = asyncio.create_task(self._scan_loop(scan_rate))

    def stop(self):
        if self._loop_task is None:
            return
        self._loop_task.cancel()

    def is_running(self):
        return not (self._loop_task is None or self._loop_task.done())



import copy
import logging
import math
from dataclasses import dataclass
from typing import AsyncGenerator

from aiospb.shared import Clock, DataType, DeviceKey, Metric, NodeOptions, UtcClock

from .ports import (
    MessageSession,
    MQTT5Client,
    MQTTError,
    MQTTMessage,
    PayloadEncoder,
    SessionIsBroken,
)
from .values import Event, Message, Seq

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class NodeMessageSession(MessageSession):
    def __init__(
        self,
        key: DeviceKey,
        mqtt_client: MQTT5Client,
        node_encoder: PayloadEncoder,
        host_encoder: PayloadEncoder,
        opts: NodeOptions | None = None,
        clock: Clock | None = None,
    ):
        self._key = key
        self._client = mqtt_client
        self._nencoder = node_encoder
        self._hencoder = host_encoder
        self._opts = opts or NodeOptions()
        self._clock = clock or UtcClock()

        self._seq = Seq(-1)
        self._bd_metric = None
        self._devices = set()

        self._connected = False

    def with_options(self, opts: NodeOptions):
        self._opts = opts

    async def establish(self):
        val = 0
        if self._bd_metric and type(self._bd_metric.value) is int:
            val = Seq(self._bd_metric.value).next().value
        ts = self._clock.timestamp()
        bd_metric = Metric(ts, val, DataType.Int64, name="bdSeq")
        msg = Message(Event.NODE_HAS_DIED, self._key, sending_ts=ts)

        await self._client.connect(
            will=MQTTMessage(
                msg.get_topic(),
                self._nencoder.encode(msg.get_payload(bd_metric=bd_metric)),
                1,
                False,
            )
        )

        # Subscribe to host messages and command messages
        await self._client.subscribe("spBv1.0/STATE/+", qos=1)
        await self._client.subscribe(
            f"spBv1.0/{self._key.group}/NCMD/{self._key.node_name}", qos=1
        )
        await self._client.subscribe(
            f"spBv1.0/{self._key.group}/DCMD/{self._key.node_name}/+", qos=1
        )

        self._bd_metric = bd_metric
        self._connected = True

    async def terminate(self):
        self._connected = False
        try:
            await self._client.disconnect()
        except MQTTError:
            pass

    def is_established(self) -> bool:
        return self._connected

    async def publish(self, msg: Message):
        if msg.event in (
            Event.HOST_STATE_HAS_CHANGED,
            Event.NODE_COMMAND_IS_SENT,
            Event.DEVICE_COMMAND_IS_SENT,
        ):
            raise ValueError(f"Message of type {msg.event} can not published")
        qos = 1 if msg.event == Event.NODE_HAS_DIED else 0
        seq = None if msg.event == Event.NODE_HAS_DIED else self._seq.next().value
        bd_metric = (
            self._bd_metric
            if msg.event in (Event.NODE_IS_BORN, Event.NODE_HAS_DIED)
            else None
        )
        msg.sending_ts = self._clock.timestamp()
        encoded = self._nencoder.encode(msg.get_payload(seq, bd_metric))

        if self._opts.max_payload_size and len(encoded) > self._opts.max_payload_size:
            if msg.event in (
                Event.DEVICE_DATA_HAS_CHANGED,
                Event.NODE_DATA_HAS_CHANGED,
            ):
                await self._publish_splitted(
                    msg, math.ceil(len(encoded) / self._opts.max_payload_size)
                )
                return
            else:
                raise ValueError(
                    f"Size of {msg.event} is over {self._opts.max_payload_size} bytes"
                )

        try:
            await self._client.publish(
                MQTTMessage(
                    msg.get_topic(),
                    encoded,
                    qos,
                    False,
                )
            )
        except MQTTError as e:
            logger.error(f"Message {msg.event} can not be send, by MQTTError")
            logger.exception(e)
            self._connected = False
            raise SessionIsBroken()

        self._seq = self._seq.next()

    async def _publish_splitted(self, msg: Message, n_parts: int):
        m_len = len(msg.metrics)
        size = m_len // n_parts

        if size == 0:
            raise ValueError(
                f"Not possible to split {m_len} metrics into {n_parts} parts"
            )

        low = 0
        high = size
        while True:
            if high + size > m_len:
                high = m_len

            split = copy.copy(msg)
            split.metrics = split.metrics[low:high]
            await self.publish(split)

            if high == m_len:
                return

            low = high
            high = high + size

    async def subscribe(self) -> AsyncGenerator[Message, None]:
        try:
            async for msg in self._client.messages:
                try:
                    if msg.topic.startswith("spBv1.0/STATE/"):
                        payload = self._hencoder.decode(msg.payload)
                    else:
                        payload = self._nencoder.decode(msg.payload)

                    result = Message.construct(msg.topic, payload)

                    if result.event not in (
                        Event.HOST_STATE_HAS_CHANGED,
                        Event.DEVICE_COMMAND_IS_SENT,
                        Event.NODE_COMMAND_IS_SENT,
                    ):
                        logger.warning(
                            f"NodeMessageSession can not recieve {result.event} messages, it will be lost"
                        )
                        continue

                    if (
                        result.event
                        in (Event.DEVICE_COMMAND_IS_SENT, Event.NODE_COMMAND_IS_SENT)
                        and result.key.node != self._key
                    ):
                        logger.warning(
                            f'Command requested to other node "{result.key.node}, it will be lost"'
                        )
                        continue

                    yield result
                except Exception as e:
                    logger.warning(
                        "Processing message from {msg.topic}, it will be lost"
                    )
                    logger.exception(e)
        except MQTTError as e:
            logger.warning("Connection to MQTT Server is broken")
            logger.exception(e)
        except Exception as e:
            logger.error("Unexpected exception from MQTT Server")
            logger.exception(e)

        self._connected = False
        raise SessionIsBroken()



from dataclasses import dataclass

from aiospb.hosts import CommandFollower, Host, MetricsCreator
from aiospb.infra.encoding import JsonPayloadEncoder, ProtobufPayloadEncoder
from aiospb.infra.fs import FSHistorian
from aiospb.infra.inmem import (
    InMemHistorian,
    InMemMinMetricInfoCache,
    InMemReadingStore,
)
from aiospb.infra.paho import MQTTConfig, PahoMQTTClient
from aiospb.messages import HostMessageSession, NodeMessageSession
from aiospb.nodes import EdgeNode, MetricsNetwork, NodeDevice
from aiospb.shared import DeviceKey, HostOptions, NodeOptions


@dataclass
class HostConfig:
    hostname: str  # it shall not have any "/"
    max_payload_size: int = 0  # unit is bytes
    reorder_time: float = 2.0  # unit is s


def new_host(
    key: str, mqtt_config: MQTTConfig, host_opts: HostOptions
) -> Host:  # pragma: no cover
    """piping to create a sparkplug host from simple parameters"""

    host_key = DeviceKey(key)
    if not host_key.hostname:
        raise ValueError(f'Not allowed name of host "{key}" with a /')
    mqtt_client = PahoMQTTClient(key, mqtt_config)

    session = HostMessageSession(
        host_key,
        mqtt_client,
        ProtobufPayloadEncoder(),
        JsonPayloadEncoder(),
        host_opts,
    )
    cache = InMemMinMetricInfoCache()
    return Host(host_key, session, MetricsCreator(cache), CommandFollower(), host_opts)


def new_edge_of_node(
    key: str,
    mqtt_config: MQTTConfig,
    node_options: NodeOptions,
    metrics_net: MetricsNetwork,
):
    node_key = DeviceKey(key)
    if node_key.hostname or node_key.device_name:
        raise ValueError(f'Not allowed name "{key}" for a node key')

    mqtt_client = PahoMQTTClient(key, mqtt_config)
    session = NodeMessageSession(
        node_key,
        mqtt_client,
        ProtobufPayloadEncoder(),
        JsonPayloadEncoder(),
        node_options,
    )
    historian = (
        FSHistorian(node_options.historian_dir)
        if node_options.historian_dir
        else InMemHistorian()
    )
    readings_store = InMemReadingStore()
    node = EdgeNode(
        NodeDevice(node_key, metrics_net, readings_store),
        session,
        historian,
        node_options,
    )

    keys = metrics_net.get_device_keys()
    for device_key in keys:
        if device_key.node != node_key:
            node.add_device(NodeDevice(device_key, metrics_net, readings_store))

    return node



import asyncio
import logging
from typing import AsyncIterable

from aiospb.messages.ports import (
    MessageSession,
    MQTT5Client,
    MQTTError,
    MQTTMessage,
    PayloadEncoder,
    SessionIsBroken,
)
from aiospb.shared import Clock, DataType, HostOptions, Metric, UtcClock

from .values import DeviceKey, Event, Message, Seq

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class _Sorter:
    def __init__(
        self,
        node_key: DeviceKey,
        output: asyncio.Queue,
        reorder_time: float = 2.0,
    ):
        self._node_key = node_key
        self._output = output
        self._time = reorder_time

        self._seq = None
        self._bd_seq = None
        self._pending = {}
        self._oos_task = None

    async def _raise_out_of_sequence(self, ts: int, seq: int, bd_seq: int):
        await asyncio.sleep(self._time)
        ts += int(self._time * 1000)
        self._output.put_nowait(
            Message(
                Event.OUT_OF_SEQUENCE,
                self._node_key,
                [
                    Metric(ts, seq, DataType.Int64, name="seq"),
                    Metric(ts, bd_seq, DataType.Int16, name="bdSeq"),
                ],
            )
        )
        self._oos_task = None

    def push_pack(self, msg: Message, seq: int | None, bd_metric: Metric | None):
        if msg.event == Event.NODE_HAS_DIED and bd_metric:
            if self._bd_seq is None or self._bd_seq != bd_metric.value:
                should = self._bd_seq.value if self._bd_seq else None
                logger.warning(
                    f'Lost of node death at "{msg.key}", should be {should} but is {bd_metric.value}'
                )
                return

        if msg.event == Event.NODE_IS_BORN and bd_metric and seq is not None:
            val = bd_metric.value if type(bd_metric.value) is int else 0
            if not self._bd_seq:
                self._bd_seq = Seq(val)
            else:
                bd_seq = self._bd_seq
                if bd_seq.value != val or bd_seq.next().value != val:
                    logger.warning(f'Incorrect bdSeq at "{msg.key}"')
                self._bd_seq = Seq(val)

            if self._pending:
                logger.warning(
                    f"Lost {len(self._pending)} messages at {self._node_key}, by rebirth"
                )
                if self._oos_task:
                    self._oos_task.cancel()
                self._pending.clear()

            self._seq = Seq(seq)
            self._output.put_nowait(msg)
            return

        if seq is not None and not bd_metric:  # Rest of SpbMessages
            if self._seq is None:
                logger.warning(
                    f'Lost {msg.__class__.__name__} by no birth of "{self._node_key}"'
                )
                return

            _seq = self._seq.next()
            if seq == _seq.value:  # It is tne next message
                self._output.put_nowait(msg)
                self._seq = _seq
                while True:
                    _seq = _seq.next()
                    msg = self._pending.pop(_seq.value, None)
                    if not msg:
                        break
                    self._output.put_nowait(msg)
                    self._seq = _seq
                if self._oos_task:
                    self._oos_task.cancel()
            else:
                self._oos_task = asyncio.create_task(
                    self._raise_out_of_sequence(
                        msg.sending_ts,
                        _seq.value,
                        self._bd_seq.value if self._bd_seq else 0,
                    )
                )
                self._pending[seq] = msg


class HostMessageSession(MessageSession):
    def __init__(
        self,
        key: DeviceKey,
        mqtt_client: MQTT5Client,
        node_encoder: PayloadEncoder,
        host_encoder: PayloadEncoder,
        opts: HostOptions,
        clock: Clock | None = None,
    ):
        if not key.hostname:
            raise ValueError(f'DeviceKey "{key}" is not a correct hostname')
        self._key = key
        self._client = mqtt_client
        self._nencoder = node_encoder
        self._hencoder = host_encoder
        self._opts = opts
        self._clock = clock or UtcClock()

        self._connected = False
        self._messages = asyncio.Queue()

        self._sorters = {}
        self._queue = asyncio.Queue()

    async def establish(self):
        try:
            await self._client.connect(will=self._host_state_is_online_to_mqtt(False))
            await self._client.subscribe(f"spBv1.0/{self._opts.group}/+/+/#", qos=1)
            await self._client.publish(self._host_state_is_online_to_mqtt(True))
            self._connected = True
        except MQTTError as e:
            logger.error("Connection broken by MQTTError")
            logger.exception(e)
            self._connected = False
            raise SessionIsBroken() from e

    def _host_state_is_online_to_mqtt(self, online: bool):
        msg = Message(
            Event.HOST_STATE_HAS_CHANGED,
            self._key,
            online=online,
            sending_ts=self._clock.timestamp(),
        )
        return MQTTMessage(
            msg.get_topic(), self._hencoder.encode(msg.get_payload()), 1, True
        )

    async def terminate(self):
        if not self._connected:
            return

        self._connected = False

        try:
            await self._client.publish(self._host_state_is_online_to_mqtt(False))
            await self._client.disconnect()
        except MQTTError:
            ...

    def is_established(self) -> bool:
        return self._connected

    async def publish(self, msg: Message):
        if msg.event not in (Event.NODE_COMMAND_IS_SENT, Event.DEVICE_COMMAND_IS_SENT):
            raise ValueError(
                f"HostMessageSession only publish commands, not {msg.event} messages"
            )

        msg.sending_ts = self._clock.timestamp()
        try:
            await self._client.publish(
                MQTTMessage(
                    msg.get_topic(), self._nencoder.encode(msg.get_payload()), 0, False
                )
            )
        except MQTTError as e:
            logger.error("Connection broken by MQTTError")
            logger.exception(e)
            self._connected = False
            self._receipt_task = None
            raise SessionIsBroken() from e

    async def subscribe(self) -> AsyncIterable[Message]:
        self._receipt_task = asyncio.create_task(self._recieve_msgs())
        try:
            while True:
                msg = await self._messages.get()
                if isinstance(msg, Exception):
                    raise msg
                yield msg
        finally:
            if self._receipt_task:
                self._receipt_task.cancel()
                self._receipt_task = None

    def _process_pack(self, msg: Message, seq: int | None, bd_metric: Metric | None):
        if msg.key not in self._sorters:
            self._sorters[msg.key] = _Sorter(
                msg.key, self._messages, self._opts.reorder_time
            )

        self._sorters[msg.key].push_pack(msg, seq, bd_metric)

    async def _recieve_msgs(self):
        try:
            async for mqtt_msg in self._client.messages:
                if "NCMD" in mqtt_msg.topic or "DCMD" in mqtt_msg.topic:
                    continue

                try:
                    payload = self._nencoder.decode(mqtt_msg.payload)
                    bd_metric = None
                    if payload.metrics:
                        for m in payload.metrics:
                            if m.name == "bdSeq":
                                bd_metric = m
                                break
                    msg = Message.construct(mqtt_msg.topic, payload)
                    self._process_pack(msg, payload.seq, bd_metric)
                except Exception as e:
                    logger.error(
                        f"Procesing message from {mqtt_msg.topic}, it will be lost"
                    )
                    logger.exception(e)
        except MQTTError as e:
            logger.error("Connection broken by MQTTError")
            logger.exception(e)
            self._connected = False
            self._receipt_task = None
            self._messages.put_nowait(SessionIsBroken())
